﻿using NiceHashMiner.Enums;
using System;
using NiceHashMiner.Switching;

namespace NiceHashMiner
{
    public class Algorithm
    {
        public readonly string AlgorithmName;
        public readonly string MinerBaseTypeName;
        public readonly AlgorithmType NiceHashID;
        public readonly AlgorithmType SecondaryNiceHashID;
        public readonly MinerBaseType MinerBaseType;

        public readonly string AlgorithmStringID;

        // Miner name is used for miner ALGO flag parameter
        public string MinerName;

        public double BenchmarkSpeed { get; set; }
        public double SecondaryBenchmarkSpeed { get; set; }
        public string ExtraLaunchParameters { get; set; }
        public bool Enabled { get; set; }

        // CPU miners only setting
        public int LessThreads { get; set; }

        // avarage speed of same devices to increase mining stability
        public double AvaragedSpeed { get; set; }

        public double SecondaryAveragedSpeed { get; set; }

        // based on device and settings here we set the miner path
        public string MinerBinaryPath = "";

        // these are changing (logging reasons)
        public double CurrentProfit = 0;

        public double CurNhmSmaDataVal = 0;
        public double SecondaryCurNhmSmaDataVal = 0;

        public Algorithm(MinerBaseType minerBaseType, AlgorithmType niceHashID, string minerName,
            AlgorithmType secondaryNiceHashID = AlgorithmType.NONE)
        {
            NiceHashID = niceHashID;
            SecondaryNiceHashID = secondaryNiceHashID;

            AlgorithmName = AlgorithmNiceHashNames.GetName(DualNiceHashID());
            MinerBaseTypeName = Enum.GetName(typeof(MinerBaseType), minerBaseType);
            AlgorithmStringID = MinerBaseTypeName + "_" + AlgorithmName;

            MinerBaseType = minerBaseType;
            MinerName = minerName;

            BenchmarkSpeed = 0.0d;
            SecondaryBenchmarkSpeed = 0.0d;
            ExtraLaunchParameters = "";
            LessThreads = 0;
            Enabled = !(NiceHashID == AlgorithmType.Nist5 ||
                        (NiceHashID == AlgorithmType.NeoScrypt && minerBaseType == MinerBaseType.sgminer));
            BenchmarkStatus = "";
        }

        // benchmark info
        public string BenchmarkStatus { get; set; }

        public bool IsBenchmarkPending { get; private set; }

        public string CurPayingRatio
        {
            get
            {
                var ratio = International.GetText("BenchmarkRatioRateN_A");
                if (NHSmaData.TryGetPaying(NiceHashID, out var paying))
                {
                    ratio = paying.ToString("F8");
                    if (IsDual() && NHSmaData.TryGetPaying(SecondaryNiceHashID, out var secondaryPaying))
                    {
                        ratio += "/" + secondaryPaying.ToString("F8");
                    }
                }
                return ratio;
            }
        }

        public string CurPayingRate
        {
            get
            {
                var rate = International.GetText("BenchmarkRatioRateN_A");
                var payingRate = 0.0d;
                if (BenchmarkSpeed > 0 && NHSmaData.TryGetPaying(NiceHashID, out var paying))
                {
                    payingRate += BenchmarkSpeed * paying * 0.000000001;
                }
                if (SecondaryBenchmarkSpeed > 0 && IsDual() && NHSmaData.TryGetPaying(NiceHashID, out var secondaryPaying))
                {
                    payingRate += SecondaryBenchmarkSpeed * secondaryPaying * 0.000000001;
                }
                rate = payingRate.ToString("F8");
                return rate;
            }
        }

        public void SetBenchmarkPending()
        {
            IsBenchmarkPending = true;
            BenchmarkStatus = International.GetText("Algorithm_Waiting_Benchmark");
        }

        public void SetBenchmarkPendingNoMsg()
        {
            IsBenchmarkPending = true;
        }

        private bool IsPendingString()
        {
            return BenchmarkStatus == International.GetText("Algorithm_Waiting_Benchmark")
                   || BenchmarkStatus == "."
                   || BenchmarkStatus == ".."
                   || BenchmarkStatus == "...";
        }

        public void ClearBenchmarkPending()
        {
            IsBenchmarkPending = false;
            if (IsPendingString())
            {
                BenchmarkStatus = "";
            }
        }

        public void ClearBenchmarkPendingFirst()
        {
            IsBenchmarkPending = false;
            BenchmarkStatus = "";
        }

        public string BenchmarkSpeedString()
        {
            if (Enabled && IsBenchmarkPending && !string.IsNullOrEmpty(BenchmarkStatus))
            {
                return BenchmarkStatus;
            }
            if (BenchmarkSpeed > 0)
            {
                return Helpers.FormatDualSpeedOutput(NiceHashID, BenchmarkSpeed, SecondaryBenchmarkSpeed);
            }
            if (!IsPendingString() && !string.IsNullOrEmpty(BenchmarkStatus))
            {
                return BenchmarkStatus;
            }
            return International.GetText("BenchmarkSpeedStringNone");
        }

        // return hybrid type if dual, else standard ID
        public AlgorithmType DualNiceHashID()
        {
            if (NiceHashID == AlgorithmType.DaggerHashimoto)
            {
                switch (SecondaryNiceHashID)
                {
                    case AlgorithmType.Decred:
                        return AlgorithmType.DaggerDecred;
                    case AlgorithmType.Lbry:
                        return AlgorithmType.DaggerLbry;
                    case AlgorithmType.Pascal:
                        return AlgorithmType.DaggerPascal;
                    case AlgorithmType.Sia:
                        return AlgorithmType.DaggerSia;
                }
            }
            return NiceHashID;
        }

        public bool IsDual()
        {
            return AlgorithmType.DaggerSia <= DualNiceHashID() && DualNiceHashID() <= AlgorithmType.DaggerPascal;
        }
    }
}
